import numpy as np
import pandas as pd
import scanpy as sc
from sklearn.cluster import KMeans
from sklearn.decomposition import PCA
from sklearn.model_selection import ParameterGrid


def select_genes(
    adata,
    n_top_genes: int,
    preselected_genes: int = None,
    seed: int = 0,
):
    """Selects genes for analysis that are either preselected or highly variable, based on a clustering heuristic.

    Parameters
    ----------
    adata :
    n_top_genes : int
        Number of genes to keep
    preselected_genes : int, optional
        Names of genes to keep, by default None
    seed : int, optional
        Seed, by default 0
    """
    adata_ = adata.copy()
    preselected_genes = preselected_genes if preselected_genes is not None else []

    adata_log = adata_.copy()
    sc.pp.normalize_total(adata_log, target_sum=1e6)
    sc.pp.log1p(adata_log)
    pca_ = PCA(n_components=50).fit(adata_log.X)
    sc.pp.highly_variable_genes(adata, n_top_genes=n_top_genes, flavor="seurat_v3")

    clusters = KMeans(n_clusters=n_top_genes, random_state=seed).fit_predict(
        pca_.components_.T
    )
    adata.var.loc[:, "clusters"] = clusters

    adata.var.index.name = "index"
    selected_genes = (
        adata.var.reset_index()
        .groupby("clusters")
        .apply(lambda x: x.sort_values("variances_norm").iloc[-1]["index"])
        .values
    )
    union_genes = np.union1d(selected_genes, preselected_genes)
    return adata[:, adata.var.index.isin(union_genes)].copy()


def one_hot(x):
    return np.eye(x.max() + 1)[x]


def select_architecture(
    model_cls,
    adata,
    xy_model_kwargs_grid: dict,
    **kwargs,
):
    """Selects architecture for feature selection"""
    all_losses = pd.DataFrame()
    parameter_grid = list(ParameterGrid(xy_model_kwargs_grid))
    keys_of_interest = None
    kwargs_ = kwargs.copy()
    if "xy_model_kwargs" not in kwargs_:
        kwargs_["xy_model_kwargs"] = {}
    for grid_param in parameter_grid:
        base_config_ = kwargs_.copy()
        base_config_["xy_model_kwargs"].update(grid_param)

        crt = model_cls(adata=adata, **base_config_)
        crt.train_statistic()
        losses = crt.losses.assign(**grid_param)
        all_losses = pd.concat([all_losses, losses])
        if keys_of_interest is None:
            keys_of_interest = list(grid_param.keys())
    relevant_losses = all_losses.loc[lambda x: x.metric == "stat_val_loss"]
    best_params = (
        relevant_losses.groupby(keys_of_interest)["value"]
        .min()
        .reset_index()
        .loc[:, keys_of_interest]
        .iloc[0]
        .to_dict()
    )
    return {**kwargs, **best_params}



import os

import fastcluster
import flax.linen as nn
import jax
import jax.numpy as jnp
import numpy as np
import pandas as pd
import plotnine as p9
import scanpy as sc
import xarray as xr
from flax import serialization
from ml_collections import ConfigDict
from scipy.cluster import hierarchy
from scipy.spatial.distance import squareform
from statsmodels.stats.multitest import multipletests
from tqdm import tqdm

from vivs._dl_utils import construct_dataloader
from vivs._models import JAXSCVAE, ImportanceScorer, ImportanceScorerLinear
from vivs._training import train_scvi, train_statistic
from vivs._utils import one_hot

from ._constants import REGISTRY_KEYS


class VIVS:
    def __init__(
        self,
        adata: sc.AnnData,
        feature_obsm_key: str = "protein_expression",
        feature_names: str = None,
        batch_key: str = None,
        n_mc_samples: int = 500,
        percent_dev: float = 0.5,
        compute_pvalue_on: str = "val",
        **config_kwargs,
    ):
        self.n_mc_samples = n_mc_samples

        self.config = self.get_default_config()
        self.config.update(config_kwargs)

        self.n_genes = adata.X.shape[1]
        self.n_features = adata.obsm[feature_obsm_key].shape[-1]
        self.n_batch = (
            adata.obs[batch_key].unique().shape[0] if batch_key is not None else 0
        )
        feature_expression = adata.obsm[feature_obsm_key]

        if batch_key is not None:
            adata.obs.loc[:, "_batch_indices"] = (
                adata.obs[batch_key].astype("category").cat.codes.values
            )
            self.batch_key = "_batch_indices"
        else:
            self.batch_key = None

        # Preprocess target expression
        self.internal_feature_key = "_vivs_feature_expression"
        self.feature_field_name = "response"
        xy_loss = self.config["xy_model_kwargs"]["loss_type"]
        if xy_loss == "mse":
            adata.obsm[self.internal_feature_key] = (
                feature_expression - feature_expression.mean(0)
            ) / feature_expression.std(0)
        elif xy_loss == "binary":
            adata.obsm[self.internal_feature_key] = feature_expression

        if feature_names is not None:
            self.feature_names = feature_names
        else:
            self.feature_names = (
                adata.obsm[feature_obsm_key].columns
                if hasattr(adata.obsm[feature_obsm_key], "columns")
                else np.arange(self.n_features)
            )
        self.compute_pvalue_on = compute_pvalue_on

        # Generative model
        self.x_tx = self.config["x_train_kwargs"]["tx"]
        self.xy_tx = self.config["xy_train_kwargs"]["tx"]
        self.x_model_kwargs = self.config["x_model_kwargs"]
        self.x_model_kwargs["n_input"] = self.n_genes

        # Feature importance model
        self.xy_include_batch = self.config["xy_include_batch_in_input"]
        self.xy_input_size = (
            self.n_genes + self.n_batch if self.xy_include_batch else self.n_genes
        )
        xy_linear = self.config["xy_linear"]
        if xy_linear:
            self.importance_cls = ImportanceScorerLinear
            self.importance_kwargs = dict(
                loss_type=self.config["xy_model_kwargs"]["loss_type"],
            )
        else:
            self.importance_cls = ImportanceScorer
            self.importance_kwargs = self.config["xy_model_kwargs"]
        self.importance_kwargs["n_features"] = self.n_features
        self.batch_size = self.config["batch_size"]
        self.n_epochs = self.config["n_epochs"]

        # Useful preprocessing
        adata_log = adata.copy()
        sc.pp.normalize_total(adata_log, target_sum=1e6)
        sc.pp.log1p(adata_log)
        self.adata_log = adata_log
        n_obs = adata.X.shape[0]
        np.random.seed(0)
        self.is_dev = np.random.random(n_obs) <= percent_dev
        self.adata = adata

        # Construct train and validation datasets
        self.adata.obs.loc[:, "is_dev"] = self.is_dev
        self.train_adata = self.adata[self.adata.obs["is_dev"]].copy()
        self.val_adata = self.adata[~self.adata.obs["is_dev"]].copy()

        # Metrics and parameters
        self.x_params = None
        self.x_batch_stats = None
        self.xy_batch_stats = None
        self.xy_params = None
        self.losses = None

    @staticmethod
    def get_default_config():
        config = ConfigDict()
        config.batch_size = 128
        config.n_epochs = 200
        config.x_model_kwargs = dict(
            n_input=None,
            n_latent=10,
            likelihood="nb",
            dropout_rate=0.0,
            n_hidden=128,
            last_h_activation="softplus",
        )
        config.x_train_kwargs = dict(
            lr=1e-3,
            patience=20,
            tx=None,
            early_stopping_metric="elbo",
            n_epochs_kl_warmup=50,
        )

        config.xy_linear = False
        config.xy_include_batch_in_input = False
        config.xy_model_kwargs = dict(
            n_features=None,
            n_hidden=128,
            dropout_rate=0.0,
            loss_type="mse",
        )

        config.xy_train_kwargs = dict(
            patience=20,
            lr=1e-3,
            tx=None,
        )
        return config.lock()

    @property
    def generative_model(self) -> nn.Module:
        return JAXSCVAE(**self.x_model_kwargs)

    @property
    def statistic_model(self) -> nn.Module:
        return self.importance_cls(**self.importance_kwargs)

    def train_scvi(self):
        """Train the generative model."""

        losses, x_params, x_batch_stats = train_scvi(
            generative_model=self.generative_model,
            train_adata=self.train_adata,
            val_adata=self.val_adata,
            batch_size=self.batch_size,
            batch_key=self.batch_key,
            protein_key=self.internal_feature_key,
            n_epochs=self.n_epochs,
            **self.config["x_train_kwargs"],
        )
        self.set_or_update_metrics(losses)
        self.x_params = x_params
        self.x_batch_stats = x_batch_stats

    def train_statistic(self):
        """Train the feature importance model."""

        losses, xy_params, xy_batch_stats = train_statistic(
            statistic_model=self.statistic_model,
            train_adata=self.train_adata,
            val_adata=self.val_adata,
            batch_size=self.batch_size,
            batch_key=self.batch_key,
            protein_key=self.internal_feature_key,
            n_epochs=self.n_epochs,
            include_batch_in_input=self.xy_include_batch,
            **self.config["xy_train_kwargs"],
        )
        self.set_or_update_metrics(losses)
        self.xy_params = xy_params
        self.xy_batch_stats = xy_batch_stats

    def train_all(self):
        self.train_scvi()
        self.train_statistic()

    def get_importance(self, eval_adata=None, batch_size=128, n_mc_per_pass=1):
        if eval_adata is None:
            if self.compute_pvalue_on == "val":
                eval_adata = self.adata[~self.adata.obs["is_dev"]].copy()
            elif self.compute_pvalue_on == "all":
                eval_adata = self.adata.copy()
        n_obs = eval_adata.X.shape[0]

        eval_dl = construct_dataloader(
            eval_adata,
            batch_size=batch_size,
            shuffle=False,
            batch_key=self.batch_key,
            protein_key=self.internal_feature_key,
        )
        total_its = n_obs // batch_size

        rng = jax.random.PRNGKey(0)
        x_rng, z_rng = jax.random.split(rng)

        tilde_ts = jnp.zeros((self.n_mc_samples, self.n_genes, self.n_features))
        observed_ts = jnp.zeros((self.n_features,))
        gene_ids = jnp.arange(self.n_genes)

        @jax.jit
        def get_tilde_t(x, batch_indices, y):
            x_ = jnp.log1p(1e6 * x / jnp.sum(x, axis=-1, keepdims=True))
            x_ = self.process_xy_input(x_, batch_indices)
            res = self.statistic_model.apply(
                {
                    "params": self.xy_params,
                    "batch_stats": self.xy_batch_stats,
                },
                x_,
                y,
                training=False,
            )["all_loss"].sum(0)
            return res

        @jax.jit
        def randomize(inputs, z_rng):
            outs = self.generative_model.apply(
                {
                    "params": self.x_params,
                    "batch_stats": self.x_batch_stats,
                },
                *inputs,
                rngs={"z": z_rng},
                training=False,
            )
            px = outs["px"]
            return px

        def _compute_loss(x, xtilde, gene_id, batch_indices, y):
            x_ = x.at[..., gene_id].set(xtilde)
            x__ = jnp.log1p(1e6 * x_ / jnp.sum(x_, axis=-1, keepdims=True))
            x__ = self.process_xy_input(x__, batch_indices)
            # shape (n_cells, n_proteins)
            res = self.statistic_model.apply(
                {
                    "params": self.xy_params,
                    "batch_stats": self.xy_batch_stats,
                },
                x__,
                y,
                training=False,
            )["all_loss"].sum(0)
            return res

        @jax.jit
        def compute_tilde_t(x, px, x_rng, batch_indices, y):
            _x_tilde = px.sample(x_rng)
            _tilde_t_k = jax.vmap(_compute_loss, (None, -1, 0, None, None), 0)(
                x, _x_tilde, gene_ids, batch_indices, y
            )
            return _tilde_t_k

        @jax.jit
        def double_compute_tilde_t(x, px, x_rng, batch_indices, y):
            _x_tilde = px.sample(x_rng, sample_shape=(n_mc_per_pass,))
            _fn = jax.vmap(_compute_loss, (None, -1, 0, None, None), 0)
            _fn = jax.vmap(_fn, (None, 0, None, None, None), 0)
            return _fn(x, _x_tilde, gene_ids, batch_indices, y)

        n_passes = self.n_mc_samples // n_mc_per_pass

        for tensors in tqdm(eval_dl, total=total_its):
            x = jnp.array(tensors[REGISTRY_KEYS.X_KEY])
            batch_indices = jnp.array(tensors[REGISTRY_KEYS.BATCH_KEY])
            protein_expression = jnp.array(tensors[self.feature_field_name])

            observed_ts += get_tilde_t(x, batch_indices, protein_expression) / n_obs
            px = randomize([x, batch_indices], z_rng)

            _tilde_t = []
            if n_mc_per_pass == 1:
                for _ in range(self.n_mc_samples):
                    _tilde_t_k = compute_tilde_t(
                        x, px, x_rng, batch_indices, protein_expression
                    )
                    _tilde_t.append(_tilde_t_k[None])
                    x_rng, _ = jax.random.split(x_rng)
            else:
                for _ in range(n_passes):
                    _tilde_t_k = double_compute_tilde_t(
                        x, px, x_rng, batch_indices, protein_expression
                    )
                    _tilde_t.append(_tilde_t_k)
                    x_rng, _ = jax.random.split(x_rng)
            _tilde_t = jnp.concatenate(_tilde_t, axis=0)
            tilde_ts += _tilde_t / n_obs
            z_rng, _ = jax.random.split(z_rng)
        pval = (1.0 + (observed_ts >= tilde_ts).sum(0)) / (1.0 + self.n_mc_samples)
        padj = np.array(
            [multipletests(_pval, method="fdr_bh")[1] for _pval in pval.T]
        ).T

        return dict(
            obs_ts=np.asarray(observed_ts),
            null_ts=np.asarray(tilde_ts),
            pvalues=np.asarray(pval),
            padj=padj,
        )

    def get_cell_scores(
        self,
        gene_ids,
        protein_ids=None,
        eval_adata=None,
        batch_size=None,
        n_mc_samples=None,
    ):
        if eval_adata is None:
            eval_adata = self.adata[~self.adata.obs["is_dev"]].copy()
        n_obs = eval_adata.X.shape[0]
        batch_size = batch_size if batch_size is not None else self.batch_size
        n_mc_samples = n_mc_samples if n_mc_samples is not None else self.n_mc_samples

        eval_dl = construct_dataloader(
            eval_adata,
            batch_size=batch_size,
            shuffle=False,
            batch_key=self.batch_key,
            protein_key=self.internal_feature_key,
        )
        total_its = n_obs // batch_size

        rng = jax.random.PRNGKey(0)
        x_rng, z_rng = jax.random.split(rng)

        gene_ids = jnp.array(gene_ids)
        protein_ids = (
            jnp.array(protein_ids)
            if protein_ids is not None
            else jnp.arange(self.n_features)
        )
        tilde_t_mean = []
        obs_t = []

        @jax.jit
        def get_tilde_t_nosum(x, batch_indices, y):
            x_ = jnp.log1p(1e6 * x / jnp.sum(x, axis=-1, keepdims=True))
            x_ = self.process_xy_input(x_, batch_indices)
            res = self.statistic_model.apply(
                {
                    "params": self.xy_params,
                    "batch_stats": self.xy_batch_stats,
                },
                x_,
                y,
                training=False,
            )["all_loss"][..., protein_ids]
            return res

        def _compute_loss(x, xtilde, gene_id, batch_indices, y):
            x_ = x.at[..., gene_id].set(xtilde)
            x__ = jnp.log1p(1e6 * x_ / jnp.sum(x_, axis=-1, keepdims=True))
            x__ = self.process_xy_input(x__, batch_indices)
            # shape (n_cells, n_proteins)
            res = self.statistic_model.apply(
                {
                    "params": self.xy_params,
                    "batch_stats": self.xy_batch_stats,
                },
                x__,
                y,
                training=False,
            )["all_loss"][..., protein_ids]
            return res

        @jax.jit
        def randomize(inputs, z_rng):
            outs = self.generative_model.apply(
                {
                    "params": self.x_params,
                    "batch_stats": self.x_batch_stats,
                },
                *inputs,
                rngs={"z": z_rng},
                training=False,
            )
            px = outs["px"]
            return px

        @jax.jit
        def compute_tilde_t(x, px, x_rng, batch_indices, y):
            _x_tilde = px.sample(x_rng)[..., gene_ids]
            _tilde_t_k = jax.vmap(_compute_loss, (None, -1, 0, None, None), 1)(
                x, _x_tilde, gene_ids, batch_indices, y
            )
            return _tilde_t_k

        for tensors in tqdm(eval_dl, total=total_its):
            x = jnp.array(tensors[REGISTRY_KEYS.X_KEY])
            batch_indices = jnp.array(tensors[REGISTRY_KEYS.BATCH_KEY])
            protein_expression = jnp.array(tensors[self.feature_field_name])
            px = randomize([x, batch_indices], z_rng)
            _tilde_t = []
            for _ in range(n_mc_samples):
                _tilde_t_k = compute_tilde_t(
                    x, px, x_rng, batch_indices, protein_expression
                )
                # _tilde_t = _tilde_t.at[mc_sample].set(_tilde_t_k)
                _tilde_t.append(_tilde_t_k[None])
                x_rng, _ = jax.random.split(x_rng)
            _tilde_t = jnp.concatenate(_tilde_t, axis=0).mean(0)
            observed_ts = get_tilde_t_nosum(x, batch_indices, protein_expression)

            z_rng, _ = jax.random.split(z_rng)
            # score = _tilde_t - observed_ts[:, None]
            tilde_t_mean.append(np.asarray(_tilde_t))
            obs_t.append(np.asarray(observed_ts[:, None]))
        tilde_t_mean = np.concatenate(tilde_t_mean)
        obs_t = np.concatenate(obs_t)
        return dict(
            tilde_t_mean=tilde_t_mean,
            obs_t=obs_t,
        )

    def save_params(self, save_path, save_adata=False):
        x_params_bytes = serialization.to_bytes(self.x_params)
        x_batch_stats_bytes = serialization.to_bytes(self.x_batch_stats)
        xy_batch_stats_bytes = serialization.to_bytes(self.xy_batch_stats)
        xy_params_bytes = serialization.to_bytes(self.xy_params)

        files = [
            (x_params_bytes, "x_params_bytes"),
            (x_batch_stats_bytes, "x_batch_stats_bytes"),
            (xy_batch_stats_bytes, "xy_batch_stats_bytes"),
            (xy_params_bytes, "xy_params_bytes"),
        ]
        for file, name in files:
            with open(os.path.join(save_path, name), "wb") as f:
                f.write(file)

        if save_adata:
            self.adata.write(os.path.join(save_path, "adata.h5ad"))

        self.init_params.to_csv(os.path.join(save_path, "init_params.csv"))

    def get_gene_correlations(self, adata=None):
        """Compute G times G gene correlation matrix."""

        cpu_device = jax.devices("cpu")[0]
        scdl = construct_dataloader(
            adata,
            batch_size=self.batch_size,
            shuffle=True,
            batch_key=self.batch_key,
            protein_key=self.internal_feature_key,
        )

        @jax.jit
        def get_scales(inputs, z_rng):
            return self.generative_model.apply(
                {
                    "params": self.x_params,
                    "batch_stats": self.x_batch_stats,
                },
                *inputs,
                rngs={"z": z_rng},
                training=False,
            )["h"]

        xx_est = jax.device_put(
            jnp.zeros((self.n_genes, self.n_genes)), device=cpu_device
        )
        x_est = jax.device_put(jnp.zeros(self.n_genes), device=cpu_device)
        z_rng = jax.random.PRNGKey(0)
        n_obs_total = adata.X.shape[0]
        for tensors in scdl:
            x = jnp.array(tensors[REGISTRY_KEYS.X_KEY])
            batch_indices = jnp.array(tensors[REGISTRY_KEYS.BATCH_KEY])
            z_rng = jax.random.PRNGKey(0)
            z_rng, _ = jax.random.split(z_rng)
            scales = get_scales([x, batch_indices], z_rng)
            x_est += jax.device_put(
                jnp.sum(scales, axis=0) / n_obs_total, device=cpu_device
            )
            xx_est += jax.device_put(
                jnp.sum(
                    jax.lax.batch_matmul(scales[..., None], scales[:, None]), axis=0
                )
                / n_obs_total,
                device=cpu_device,
            )
        x_est = x_est[None]
        cov_ = xx_est - jnp.matmul(x_est.T, x_est)
        factor_ = 1.0 / jnp.sqrt(jnp.diag(cov_))
        dmat = jnp.diag(factor_)
        corr_ = dmat @ (cov_ @ dmat)
        return np.asarray(corr_)

    def get_gene_groupings(
        self,
        adata=None,
        method="complete",
        return_z=False,
        n_clusters_list=None,
    ):
        """Computes gene groupings based on gene correlations.

        Parameters
        ----------
        adata :
            adata used to compute gene correlations.
        method :
            Linkage for hierarchical clustering.
        return_z :
            Whether to return linkage matrix.
        n_clusters_list :
            Number of desired clusters.
        """
        adata = self.adata[self.adata.obs.is_dev].copy() if adata is None else adata
        corr_ = self.get_gene_correlations(adata=adata)
        pseudo_dist = 1 - corr_
        pseudo_dist = (pseudo_dist + pseudo_dist.T) / 2
        pseudo_dist = np.clip(pseudo_dist, a_min=0.0, a_max=100.0)
        pseudo_dist = pseudo_dist - np.diag(np.diag(pseudo_dist))
        dist_vec = squareform(pseudo_dist, checks=False)
        Z = fastcluster.linkage(dist_vec, method=method)
        Z = hierarchy.optimal_leaf_ordering(Z, dist_vec)
        gene_order = hierarchy.leaves_list(Z)
        gene_order = self.adata.var_names[gene_order].values

        assert n_clusters_list is not None
        if not isinstance(n_clusters_list, list):
            n_clusters_list = [n_clusters_list]
        gene_groupings = []
        for n_cluster in n_clusters_list:
            if n_cluster >= self.n_genes:
                continue
            cluster_assignments = hierarchy.fcluster(Z, n_cluster, criterion="maxclust")
            cluster_assignments -= 1
            gene_groupings.append(cluster_assignments)
        if return_z:
            return (
                gene_groupings,
                Z,
                gene_order,
            )
        return gene_groupings

    def predict_t(self, adata=None, batch_size=128):
        @jax.jit
        def get_t(x, y):
            x_ = jnp.log1p(1e6 * x / jnp.sum(x, axis=-1, keepdims=True))
            res = self.statistic_model.apply(
                {
                    "params": self.xy_params,
                    "batch_stats": self.xy_batch_stats,
                },
                x_,
                y,
                training=False,
            )["all_loss"]
            return res

        adata = self.adata if adata is None else adata
        eval_dl = construct_dataloader(
            adata,
            batch_size=batch_size,
            shuffle=False,
            batch_key=self.batch_key,
            protein_key=self.internal_feature_key,
        )
        res = []
        for tensors in eval_dl:
            x = jnp.array(tensors[REGISTRY_KEYS.X_KEY])
            # batch_indices = jnp.array(tensors[REGISTRY_KEYS.BATCH_KEY])
            protein_expression = jnp.array(tensors[self.feature_field_name])
            res.append(np.array(get_t(x, protein_expression)))
        return np.concatenate(res, axis=0)

    def get_latent(self):
        dl = construct_dataloader(
            self.adata,
            batch_size=128,
            shuffle=False,
            batch_key=self.batch_key,
            protein_key=self.internal_feature_key,
        )

        @jax.jit
        def _get_latent(inputs, z_rng):
            outs = self.generative_model.apply(
                {
                    "params": self.x_params,
                    "batch_stats": self.x_batch_stats,
                },
                *inputs,
                rngs={"z": z_rng},
                training=False,
            )
            return outs["qz"].loc

        cpu_device = jax.devices("cpu")[0]
        zs = []
        z_rng = jax.random.PRNGKey(0)
        for tensors in dl:
            x = jnp.array(tensors[REGISTRY_KEYS.X_KEY])
            batch_indices = jnp.array(tensors[REGISTRY_KEYS.BATCH_KEY])
            qz = _get_latent([x, batch_indices], z_rng)
            qz = jax.device_put(qz, device=cpu_device)
            zs.append(qz)
        zs = jnp.concatenate(zs, axis=0)
        return np.array(zs)

    def get_hier_importance(
        self,
        n_clusters_list: list,
        eval_adata: sc.AnnData = None,
        batch_size=128,
        gene_groupings=None,
        gene_order=None,
        clustering_method="complete",
        use_vmap=True,
    ):
        """
        Performs hierarchical importance testing.

        This method first computes groups of genes based on their correlations, or alternatively,
        relies on pre-computed gene groupings at several resolutions.
        At each resolution, conditional independence testing is performed using the CRT.

        Once this method has been called, the results can be visualized using the `plot_hier_importance` method.

        This method returns three objects:
        - `gene_results`: an xarray object containing significance scores for each gene and protein.
        - `cluster_results` and `gene_to_cluster`: dataframes that are mostly useful for visualization,
        and can be ignored in most cases.

        Parameters
        ----------
        cluster_sizes
            List of cluster sizes, defining the resolutions at which to perform conditional independence testing.
            For instance, if cluster_sizes = [5, 10], then the method will
            test for conditional independence based on a partition of the genes into 5, 10 clusters, as well as all genes.
        eval_adata
            Anndata object to evaluate on. If None, then the validation set is used.
        batch_size
            Batch size for evaluation.
        gene_groupings
            Optional gene groupings. If None, then gene groupings are computed using hierarchical clustering.
        gene_order
            Optional gene order. This only determines how the genes are ordered in the visualization.
            If None, then gene order is computed using hierarchical clustering.
        clustering_method
            Linkage method for hierarchical clustering.
        use_vmap
            Whether to use vmap for parallelization.
        """
        if eval_adata is None:
            if self.compute_pvalue_on == "val":
                eval_adata = self.adata[~self.adata.obs["is_dev"]].copy()
            elif self.compute_pvalue_on == "all":
                eval_adata = self.adata.copy()
        n_obs = eval_adata.X.shape[0]

        # construct dataloader
        eval_dl = construct_dataloader(
            eval_adata,
            batch_size=batch_size,
            shuffle=False,
            batch_key=self.batch_key,
            protein_key=self.internal_feature_key,
        )
        total_its = n_obs // batch_size

        # compute gene groups
        train_adata = self.adata[self.adata.obs["is_dev"]].copy()
        if gene_groupings is None:
            gene_groupings, _, gene_order = self.get_gene_groupings(
                adata=train_adata,
                n_clusters_list=n_clusters_list,
                return_z=True,
                method=clustering_method,
            )
        else:
            if gene_order is None:
                gene_order = np.arange(self.n_genes)
        gene_groupings.append(np.arange(self.n_genes).astype(np.int32))
        gene_groups_oh = [jnp.array(one_hot(grouping)) for grouping in gene_groupings]
        gene_groups_sizes = [gg.shape[1] for gg in gene_groups_oh]

        rng = jax.random.PRNGKey(0)
        x_rng, z_rng = jax.random.split(rng)

        tilde_ts = [
            jnp.zeros((self.n_mc_samples, sz, self.n_features))
            for sz in gene_groups_sizes
        ]
        observed_ts = jnp.zeros((self.n_features,))

        @jax.jit
        def get_tilde_t(x, y):
            x_ = jnp.log1p(1e6 * x / jnp.sum(x, axis=-1, keepdims=True))
            res = self.statistic_model.apply(
                {
                    "params": self.xy_params,
                    "batch_stats": self.xy_batch_stats,
                },
                x_,
                y,
                training=False,
            )["all_loss"].sum(0)
            return res

        def randomize(inputs, z_rng):
            outs = self.generative_model.apply(
                {
                    "params": self.x_params,
                    "batch_stats": self.x_batch_stats,
                },
                *inputs,
                rngs={"z": z_rng},
                training=False,
            )
            px = outs["px"]
            return px

        def _compute_loss_group(x, xtilde, gene_groups, y):
            x_ = x * (1.0 - gene_groups) + (xtilde * gene_groups)
            x__ = jnp.log1p(1e6 * x_ / jnp.sum(x_, axis=-1, keepdims=True))
            # shape (n_cells, n_proteins)
            res = self.statistic_model.apply(
                {
                    "params": self.xy_params,
                    "batch_stats": self.xy_batch_stats,
                },
                x__,
                y,
                training=False,
            )["all_loss"].sum(0)
            return res

        @jax.jit
        def compute_tilde_t_group(x, x_tilde, y, gene_groups):
            if use_vmap:
                _tilde_t_k = jax.vmap(_compute_loss_group, (None, None, -1, None), 0)(
                    x, x_tilde, gene_groups, y
                )
            else:

                def parallel_fn(gene_group):
                    return _compute_loss_group(x, x_tilde, gene_group, y)

                _tilde_t_k = jax.lax.map(parallel_fn, gene_groups)
            return _tilde_t_k

        for tensors in tqdm(eval_dl, total=total_its):
            x = jnp.array(tensors[REGISTRY_KEYS.X_KEY])
            batch_indices = jnp.array(tensors[REGISTRY_KEYS.BATCH_KEY])
            protein_expression = jnp.array(tensors[self.feature_field_name])

            observed_ts += get_tilde_t(x, protein_expression) / n_obs
            px = randomize([x, batch_indices], z_rng)

            for k in range(self.n_mc_samples):
                x_tilde = px.sample(x_rng)
                x_rng, _ = jax.random.split(x_rng)
                for gene_group_idx, gene_group_oh in enumerate(gene_groups_oh):
                    _tilde_t_k = (
                        compute_tilde_t_group(
                            x, x_tilde, protein_expression, gene_group_oh
                        )
                        / n_obs
                    )

                    updated_tilde_t = tilde_ts[gene_group_idx][k] + _tilde_t_k
                    tilde_ts[gene_group_idx] = (
                        tilde_ts[gene_group_idx].at[k].set(updated_tilde_t)
                    )
            z_rng, _ = jax.random.split(z_rng)

        gene_results = self._construct_results(
            observed_ts=observed_ts,
            tilde_ts=tilde_ts,
            gene_groupings=gene_groupings,
            gene_groups_sizes=gene_groups_sizes,
            gene_order=gene_order,
        )
        return gene_results.squeeze()

    @staticmethod
    def plot_hier_importance(
        gene_results,
        base_resolution=None,
        significance_threshold=0.1,
        plot_fig=True,
        theme_kwargs=None,
    ):
        if base_resolution is None:
            base_resolution = gene_results.resolution.values.min()
        padjs = gene_results.loc[{"resolution": base_resolution}]["padj"].to_pandas()
        genes_to_plot = padjs.loc[lambda x: x < significance_threshold].index.tolist()
        res_subset = gene_results.loc[{"gene_name": genes_to_plot}].assign(
            gene_index=("gene_name", np.arange(len(genes_to_plot)))
        )
        plot_df = []
        for resolution_idx, resolution in enumerate(res_subset.resolution.values):
            res_ = res_subset.loc[{"resolution": resolution}]
            unique_clusters = np.unique(res_["cluster_assignment"].values)
            for cluster in unique_clusters:
                gene_is_in_cluster = res_["cluster_assignment"] == cluster
                res_cluster = res_.loc[{"gene_name": gene_is_in_cluster}]
                xmin = res_cluster.gene_index.values.min()
                xmax = res_cluster.gene_index.values.max()
                are_indices_contiguous = (
                    xmax - xmin + 1 == res_cluster.gene_name.shape[0]
                )
                if not are_indices_contiguous:
                    raise ValueError("Gene indices are not contiguous")
                is_cluster_detected = (res_cluster["padj"] < 0.1).all()
                if is_cluster_detected:
                    plot_df.append(
                        {
                            "resolution_idx": resolution_idx,
                            "resolution": resolution,
                            "xmin": xmin - 0.5,
                            "xmax": xmax + 0.5,
                            "ymin": resolution_idx,
                            "ymax": resolution_idx + 1,
                        }
                    )
        plot_df = pd.DataFrame(plot_df)
        labels = list(res_subset.gene_name.values)
        breaks = list(res_subset.gene_index.values)
        if plot_fig:
            if theme_kwargs is None:
                theme_kwargs = dict(figure_size=(15, 2))
            fig = (
                p9.ggplot(plot_df)
                + p9.geom_rect(
                    plot_df,
                    p9.aes(xmin="xmin", xmax="xmax", ymin="ymin", ymax="ymax"),
                    inherit_aes=False,
                    fill="#ededed",
                    color="#080808",
                )
                + p9.scale_x_continuous(labels=labels, breaks=breaks)
                + p9.theme_classic()
                + p9.theme(
                    axis_text_x=p9.element_text(rotation=90),
                    axis_line_y=p9.element_blank(),
                    axis_text_y=p9.element_blank(),
                    axis_ticks_major_y=p9.element_blank(),
                )
                + p9.labs(x="", y="")
                + p9.theme(**theme_kwargs)
            )
            return fig
        else:
            return plot_df, labels, breaks

    def _construct_results(
        self,
        observed_ts: jnp.ndarray,
        tilde_ts: jnp.ndarray,
        gene_groupings: list,
        gene_groups_sizes: list,
        gene_order: list,
    ):
        """
        Constructs the results for the hierarchical importance model based on the
        null and randomized statistics and the gene groupings.

        Parameters
        ----------
        observed_ts
            The observed statistics.
        tilde_ts
            The randomized statistics.
        gene_groupings
            The gene groupings, of shape (n_resolutions, n_genes).
        gene_groups_sizes
            The total number of gene groups per resolution, of shape (n_resolutions,).
        """
        datasets = []
        for resolution_idx, resolution in enumerate(gene_groups_sizes):
            pvals = (1.0 + (observed_ts >= tilde_ts[resolution_idx]).sum(0)) / (
                1.0 + self.n_mc_samples
            )

            padjs = [
                multipletests(pvals[:, protein_id], method="fdr_bh")[1]
                for protein_id in range(self.n_features)
            ]
            padjs = jnp.stack(padjs, axis=-1)
            gene_clusters = gene_groupings[resolution_idx]
            pvals_ = pvals[gene_clusters]
            padjs_ = padjs[gene_clusters]
            coords_ = dict(
                dims=["gene_name", "feature"],
                coords={
                    "gene_name": self.adata.var_names.values,
                    "feature": self.feature_names,
                    "resolution": resolution,
                    "resolution_idx": resolution_idx,
                },
            )
            pvals_ = xr.DataArray(pvals_, **coords_)
            padjs_ = xr.DataArray(padjs_, **coords_)
            cluster_assignments = xr.DataArray(
                gene_clusters,
                dims=["gene_name"],
                coords={
                    "gene_name": self.adata.var_names.values,
                    "resolution": resolution,
                    "resolution_idx": resolution_idx,
                },
            )
            datasets.append(
                xr.Dataset(
                    {
                        "pval": pvals_,
                        "padj": padjs_,
                        "cluster_assignment": cluster_assignments,
                    }
                )
            )
        return xr.concat(datasets, dim="resolution").reindex(gene_name=gene_order)

    def set_or_update_metrics(self, metrics: pd.DataFrame):
        """
        Update the metrics dataframe.
        """

        if self.losses is None:
            self.losses = metrics
        else:
            self.losses = pd.concat([self.losses, metrics])

    def process_xy_input(
        self, x: jnp.ndarray, batch_indices: jnp.ndarray
    ) -> jnp.ndarray:
        """
        Process inputs for the feature importance model.
        """
        return (
            jnp.concatenate(
                [x, jax.nn.one_hot(batch_indices.squeeze(-1), self.n_batch)], axis=-1
            )
            if self.xy_include_batch
            else x
        )



import flax.linen as nn
import jax
import jax.numpy as jnp
import jax.random as random
import numpyro.distributions as dist
from numpyro.distributions import constraints
from numpyro.distributions.util import is_prng_key


class ZINB(dist.NegativeBinomial2):
    """Custom ZINB distribution."""

    arg_constraints = {
        "mean": constraints.positive,
        "concentration": constraints.positive,
        "gate": constraints.unit_interval,
    }
    support = constraints.nonnegative_integer

    def __init__(self, mean, concentration, gate, *, validate_args=None):
        self.gate = gate
        super(ZINB, self).__init__(mean, concentration, validate_args=validate_args)

    def sample(self, key, sample_shape=()):
        assert is_prng_key(key)
        key_bern, key_base = random.split(key)
        shape = sample_shape + self.batch_shape

        samples = super(ZINB, self).sample(key_base, sample_shape=sample_shape)
        mask = random.bernoulli(key_bern, self.gate, shape)
        return jnp.where(mask, 0, samples)

    def log_prob(self, value):
        log_prob = super(ZINB, self).log_prob(value)
        log_prob = jnp.log1p(-self.gate) + log_prob
        return jnp.where(value == 0, jnp.log(self.gate + jnp.exp(log_prob)), log_prob)


class FlaxEncoder(nn.Module):
    """Encoder for Jax VAE."""

    n_input: int
    n_latent: int
    n_hidden: int
    precision: str

    def setup(self):
        """Setup encoder."""
        self.dense1 = nn.Dense(self.n_hidden, precision=self.precision)
        self.dense2 = nn.Dense(self.n_hidden, precision=self.precision)
        self.dense3 = nn.Dense(self.n_latent, precision=self.precision)
        self.dense4 = nn.Dense(self.n_latent, precision=self.precision)

        self.norm1 = nn.BatchNorm(momentum=0.99, epsilon=0.001)
        self.norm2 = nn.BatchNorm(momentum=0.99, epsilon=0.001)

    def __call__(self, x: jnp.ndarray, training: bool = False):
        """Forward pass."""
        is_eval = not training

        x_ = jnp.log1p(x)

        h = self.dense1(x_)
        h = self.norm1(h, use_running_average=is_eval)
        h = nn.relu(h)
        h = self.norm2(h, use_running_average=is_eval)
        h = nn.relu(h)

        mean = self.dense3(h)
        log_var = self.dense4(h)
        return dist.Normal(mean, nn.softplus(log_var))


class FlaxDecoder(nn.Module):
    """Decoder for Jax VAE."""

    n_input: int
    n_hidden: int
    precision: str
    last_h_activation: str = "softmax"

    def setup(self):
        """Setup decoder."""
        self.dense1 = nn.Dense(self.n_hidden, precision=self.precision)
        self.dense2 = nn.Dense(self.n_hidden, precision=self.precision)
        self.dense3 = nn.Dense(self.n_hidden, precision=self.precision)
        self.dense4 = nn.Dense(self.n_hidden, precision=self.precision)
        self.dense5 = nn.Dense(self.n_input, precision=self.precision)

        self.norm1 = nn.BatchNorm(momentum=0.99, epsilon=0.001)
        self.norm2 = nn.BatchNorm(momentum=0.99, epsilon=0.001)
        self.disp = self.param(
            "disp", lambda rng, shape: jax.random.normal(rng, shape), (self.n_input, 1)
        )

        self.zi_logits1 = nn.Dense(self.n_hidden)
        self.zi_logits2 = nn.Dense(self.n_hidden)
        self.zi_logits3 = nn.Dense(self.n_input)
        self.zi_logits_norm = nn.BatchNorm(momentum=0.99, epsilon=0.001)

    def __call__(self, z: jnp.ndarray, batch: jnp.ndarray, training: bool = False):
        """Forward pass."""
        is_eval = not training

        h = self.dense1(z)
        h += self.dense2(batch)

        h = self.norm1(h, use_running_average=is_eval)
        h = nn.relu(h)
        h = self.dense3(h)
        h = self.norm2(h, use_running_average=is_eval)
        h = nn.relu(h)
        h = self.dense5(h)
        if self.last_h_activation == "softmax":
            h = nn.softmax(h)
        elif self.last_h_activation == "softplus":
            h = nn.softplus(h)

        logits = self.zi_logits1(z)
        logits += self.zi_logits2(batch)
        logits = self.zi_logits_norm(logits, use_running_average=is_eval)
        logits = nn.relu(logits)
        logits = self.zi_logits3(logits)
        probs = nn.sigmoid(logits)
        return h, self.disp.ravel(), probs


class JAXSCVAE(nn.Module):
    n_input: int
    n_latent: int
    n_hidden: int
    precision: str = None
    likelihood: str = "nb"
    dropout_rate: float = 0.0
    last_h_activation: str = "softmax"

    def setup(self):
        self.encoder = FlaxEncoder(
            self.n_input, self.n_latent, self.n_hidden, precision=self.precision
        )
        self.decoder = FlaxDecoder(
            self.n_input,
            self.n_hidden,
            precision=self.precision,
            last_h_activation=self.last_h_activation,
        )
        self.dropout = nn.Dropout(rate=self.dropout_rate)

    def __call__(
        self,
        x,
        batch_indices,
        n_samples=1,
        training: bool = False,
        use_prior=False,
        kl_weight=1.0,
    ):
        z_rng = self.make_rng("z")
        sample_shape = () if n_samples == 1 else (n_samples,)
        x_ = jnp.log1p(x)
        x_ = self.dropout(x_, deterministic=not training)
        if use_prior:
            qz = dist.Normal(0, 1)
        else:
            qz = self.encoder(x_, training=training)
        z = qz.rsample(z_rng, sample_shape=sample_shape)

        h, disp, probs = self.decoder(z, batch_indices, training=training)
        if self.last_h_activation == "softmax":
            library = x.sum(-1, keepdims=True)
            scale = h * library
        else:
            scale = h
        if self.likelihood == "nb":
            px = dist.NegativeBinomial2(scale, jnp.exp(disp))
        elif self.likelihood == "zinb":
            px = ZINB(scale, jnp.exp(disp), probs)
        else:
            px = dist.Poisson(scale)
        log_px = px.log_prob(x).sum(-1)
        kl = dist.kl_divergence(qz, dist.Normal(0, 1)).sum(-1)
        elbo = log_px - (kl_weight * kl)
        loss = -elbo.mean()
        reconstruction_loss = -log_px.mean()
        return dict(
            loss=loss, h=h, z=z, px=px, reconstruction_loss=reconstruction_loss, qz=qz
        )


class ImportanceScorer(nn.Module):
    """
    Importance score relying on a neural network.

    The employed score corresponds to the negative log likelihood
    whose parameters are predicted by a neural network.
    """

    n_hidden: int
    n_features: int
    dropout_rate: float
    loss_type: str = "mse"

    def setup(self):
        self.dense1 = nn.Dense(features=self.n_hidden)
        self.dense_res = nn.Dense(features=self.n_features)
        self.norm1 = nn.BatchNorm(momentum=0.99, epsilon=0.001)
        self.dropout1 = nn.Dropout(rate=self.dropout_rate)
        self.dense3 = nn.Dense(features=self.n_features)

        self.log_std = 0.0

    def __call__(self, x, y, training: bool = False):
        is_eval = not training

        h = self.dense1(x)
        h = self.norm1(h, use_running_average=is_eval)
        h = nn.leaky_relu(h)
        h = self.dropout1(h, deterministic=is_eval)

        h = self.dense3(h)

        if self.loss_type == "mse":
            all_loss = -dist.Normal(h, jnp.exp(self.log_std)).log_prob(y)
        elif self.loss_type == "binary":
            all_loss = -dist.Bernoulli(logits=h).log_prob(y)
        loss = all_loss.mean()
        return dict(h=h, loss=loss, all_loss=all_loss)


class ImportanceScorerLinear(nn.Module):
    """Importance score relying on a linear model."""

    n_features: int
    loss_type: str = "mse"

    def setup(self):
        self.dense1 = nn.Dense(features=self.n_features)
        self.norm1 = nn.BatchNorm(momentum=0.99, epsilon=0.001)
        self.dropout1 = nn.Dropout(rate=0.0)
        self.log_std = 0.0

    def __call__(self, x, y, training: bool = False):
        is_eval = not training
        h = self.norm1(x, use_running_average=is_eval)
        h = self.dropout1(h, deterministic=is_eval)
        h = self.dense1(h)
        if self.loss_type == "mse":
            all_loss = -dist.Normal(h, jnp.exp(self.log_std)).log_prob(y)
        elif self.loss_type == "binary":
            all_loss = -dist.Bernoulli(logits=h).log_prob(y)
        loss = all_loss.mean()
        return dict(h=h, loss=loss, all_loss=all_loss)



import numpy as np
import scipy.sparse as sp
from anndata import AnnData
from torch.utils.data import DataLoader, Dataset

from ._constants import REGISTRY_KEYS


class AnnDataset(Dataset):
    def __init__(
        self,
        adata: AnnData,
        response_key: str,
        batch_key: str = None,
    ):
        self.adata = adata

        self.x_mat = adata.X
        self.is_sparse = sp.issparse(self.x_mat)
        if batch_key is None:
            self.batch_indices = np.zeros((adata.n_obs, 1))
        else:
            # Batch indices should already be integers
            batch_indices = adata.obs[batch_key].values
            self.batch_indices = batch_indices[:, None]

        self.responses = adata.obsm[response_key]

    def __len__(self):
        return self.adata.n_obs

    def __getitem__(self, idx):
        xmat = self.x_mat[idx]
        if sp.issparse(xmat):
            xmat = xmat.toarray()
        return {
            REGISTRY_KEYS.X_KEY: xmat,
            REGISTRY_KEYS.BATCH_KEY: self.batch_indices[idx],
            "response": self.responses[idx],
        }


def construct_dataloader(
    adata: AnnData,
    batch_size: int = 128,
    shuffle: bool = False,
    batch_key: str = None,
    protein_key: str = None,
):
    # Temporary fix while scvi-tools doesn't support jax
    dataset = AnnDataset(
        adata,
        response_key=protein_key,
        batch_key=batch_key,
    )
    dataloader = DataLoader(dataset, batch_size=batch_size, shuffle=shuffle)
    return dataloader



from copy import copy

import jax
import jax.numpy as jnp
import numpy as np
import optax
import pandas as pd
from flax.training import train_state
from tqdm import tqdm

from vivs._dl_utils import construct_dataloader

from ._constants import REGISTRY_KEYS

INCH_TO_CM = 1 / 2.54


def init_generative_model(
    generative_model,
    batch_size,
    n_genes,
):
    """Initializes the generative model."""

    init_rngs = {
        "params": jax.random.PRNGKey(0),
        "z": jax.random.PRNGKey(1),
        "dropout": jax.random.PRNGKey(2),
    }
    init_counts = jnp.ones((batch_size, n_genes))
    init_batches = jnp.zeros((batch_size, 1))

    return generative_model.init(
        init_rngs,
        init_counts,
        init_batches,
    )


def train_scvi(
    generative_model,
    train_adata,
    val_adata,
    batch_size,
    batch_key,
    protein_key,
    lr,
    n_epochs,
    n_epochs_kl_warmup,
    early_stopping_metric,
    patience,
    tx=None,
):
    """Trains the generative model."""

    # Construct dataloader
    n_genes = train_adata.shape[1]
    tx = optax.adam(lr) if tx is None else tx

    train_dl = construct_dataloader(
        train_adata,
        batch_size=batch_size,
        shuffle=True,
        batch_key=batch_key,
        protein_key=protein_key,
    )
    val_dl = construct_dataloader(
        val_adata,
        batch_size=batch_size,
        shuffle=False,
        batch_key=batch_key,
        protein_key=protein_key,
    )

    # Init state
    z_rng = jax.random.PRNGKey(0)
    d_rng = jax.random.PRNGKey(1)

    variables = init_generative_model(
        generative_model,
        batch_size=batch_size,
        n_genes=n_genes,
    )
    params = variables["params"]
    batch_stats = variables["batch_stats"]

    state = train_state.TrainState.create(
        apply_fn=generative_model.apply,
        params=params,
        tx=tx,
    )

    @jax.jit
    def train_step(state, batch_stats, inputs, z_rng, d_rng, kl_weight):
        def loss_fn(params):
            loss, variables = generative_model.apply(
                {"params": params, "batch_stats": batch_stats},
                *inputs,
                rngs={"z": z_rng, "dropout": d_rng},
                mutable=["batch_stats"],
                kl_weight=kl_weight,
                training=True,
            )
            loss = loss["loss"]
            return loss, variables

        (loss, batch_stats), grads = jax.value_and_grad(loss_fn, has_aux=True)(
            state.params
        )
        batch_stats = batch_stats["batch_stats"]
        return loss, state.apply_gradients(grads=grads), batch_stats

    @jax.jit
    def evaluate(state, batch_stats, inputs, z_rng):
        outs = generative_model.apply(
            {"params": state.params, "batch_stats": batch_stats},
            *inputs,
            rngs={"z": z_rng},
            training=False,
        )
        reco_loss = outs["reconstruction_loss"]
        elbo = outs["loss"]
        return reco_loss, elbo

    best_val_loss = np.inf
    best_x_params = copy(state.params)
    best_x_batch_stats = copy(batch_stats)
    hasnt_improved = 0
    losses = []
    for epoch in tqdm(range(n_epochs)):
        kl_weight = (
            min(1.0, epoch / n_epochs_kl_warmup) if n_epochs_kl_warmup > 0 else 1.0
        )
        for batch in train_dl:
            X = batch[REGISTRY_KEYS.X_KEY]
            batch_indices = batch[REGISTRY_KEYS.BATCH_KEY]
            X = jnp.array(X)
            batch_indices = jnp.array(batch_indices)

            z_rng, _ = jax.random.split(z_rng)
            d_rng, _ = jax.random.split(d_rng)
            loss, state, batch_stats = train_step(
                state,
                batch_stats,
                [X, batch_indices],
                z_rng,
                d_rng,
                kl_weight=kl_weight,
            )

        val_reco = 0
        val_elbo = 0
        n_batches = 0
        for batch in val_dl:
            X = batch[REGISTRY_KEYS.X_KEY]
            batch_indices = batch[REGISTRY_KEYS.BATCH_KEY]
            X = jnp.array(X)
            batch_indices = jnp.array(batch_indices)
            z_rng, _ = jax.random.split(z_rng)
            _val_reco, _val_elbo = evaluate(
                state, batch_stats, [X, batch_indices], z_rng
            )
            n_batches += 1
            val_reco += _val_reco
            val_elbo += _val_elbo

        losses.append(
            dict(
                metric="train_loss",
                value=loss.item(),
                iterate=epoch,
            )
        )
        losses.append(
            dict(
                metric="kl_weight",
                value=kl_weight,
                iterate=epoch,
            )
        )
        losses.append(
            dict(
                metric="val_reco",
                value=val_reco.item() / n_batches,
                iterate=epoch,
            )
        )
        losses.append(
            dict(
                metric="val_elbo",
                value=val_elbo.item() / n_batches,
                iterate=epoch,
            )
        )

        curr_loss = val_elbo if early_stopping_metric == "elbo" else val_reco
        if curr_loss < best_val_loss:
            best_val_loss = curr_loss
            hasnt_improved = 0
            best_x_params = copy(state.params)
            best_x_batch_stats = copy(batch_stats)
        else:
            hasnt_improved += 1
        if hasnt_improved > patience:
            break

    losses = pd.DataFrame(losses)
    return losses, best_x_params, best_x_batch_stats


def init_statistic_model(statistic_model, batch_size, n_genes, n_proteins):
    """Initializes the feature importance model."""

    init_rngs = {"params": jax.random.PRNGKey(0), "dropout": jax.random.PRNGKey(1)}
    init_log_counts = jnp.ones((batch_size, n_genes))
    init_prots = jnp.ones((batch_size, n_proteins))
    return statistic_model.init(init_rngs, init_log_counts, init_prots)


def train_statistic(
    statistic_model,
    train_adata,
    val_adata,
    batch_size,
    batch_key,
    protein_key,
    lr,
    n_epochs,
    patience,
    tx=None,
    normalize_x=True,
    include_batch_in_input=False,
):
    """Trains the feature importance model."""

    params = None
    batch_stats = None
    n_genes = train_adata.shape[1]
    n_proteins = train_adata.obsm[protein_key].shape[1]
    n_samples = 0 if batch_key is None else train_adata.obs[batch_key].nunique()

    tx = optax.adam(lr) if tx is None else tx

    train_dl = construct_dataloader(
        train_adata,
        batch_size=batch_size,
        shuffle=True,
        batch_key=batch_key,
        protein_key=protein_key,
    )
    val_dl = construct_dataloader(
        val_adata,
        batch_size=batch_size,
        shuffle=False,
        batch_key=batch_key,
        protein_key=protein_key,
    )

    d_rng = jax.random.PRNGKey(1)
    n_features = n_genes + n_samples if include_batch_in_input else n_genes
    variables = init_statistic_model(
        statistic_model, batch_size, n_features, n_proteins
    )
    params = variables["params"]
    batch_stats = variables["batch_stats"]

    state = train_state.TrainState.create(
        apply_fn=statistic_model.apply,
        params=params,
        tx=tx,
    )

    @jax.jit
    def train_step(state, batch_stats, inputs, dropout_rng):
        def loss_fn(params, batch_stats):
            x, batch_sample, y = inputs
            x_ = (
                jnp.log1p(1e6 * x / jnp.sum(x, axis=-1, keepdims=True))
                if normalize_x
                else x
            )
            x_ = (
                jnp.concatenate(
                    [x_, jax.nn.one_hot(batch_sample.squeeze(-1), n_samples)], axis=-1
                )
                if include_batch_in_input
                else x_
            )
            loss, variables = statistic_model.apply(
                {"params": params, "batch_stats": batch_stats},
                x_,
                y,
                rngs={"dropout": dropout_rng},
                mutable=["batch_stats"],
                training=True,
            )
            loss = loss["loss"]
            batch_stats = variables["batch_stats"]
            return loss, batch_stats

        (loss, batch_stats), grads = jax.value_and_grad(loss_fn, has_aux=True)(
            state.params, batch_stats
        )
        return loss, state.apply_gradients(grads=grads), batch_stats

    @jax.jit
    def evaluate(state, batch_stats, inputs):
        x, batch_sample, y = inputs
        x_ = (
            jnp.log1p(1e6 * x / jnp.sum(x, axis=-1, keepdims=True))
            if normalize_x
            else x
        )
        x_ = (
            jnp.concatenate(
                [x_, jax.nn.one_hot(batch_sample.squeeze(-1), n_samples)], axis=-1
            )
            if include_batch_in_input
            else x_
        )
        loss = statistic_model.apply(
            {"params": state.params, "batch_stats": batch_stats},
            x_,
            y,
            training=False,
        )["loss"]
        return loss

    best_val_loss = np.inf
    best_params = copy(state.params)
    best_batch_stats = copy(batch_stats)
    hasnt_improved = 0
    losses = []
    for epoch in tqdm(range(n_epochs)):
        for batch in train_dl:
            X = batch[REGISTRY_KEYS.X_KEY]
            protein_exp = batch["response"]
            sample_batch = batch[REGISTRY_KEYS.BATCH_KEY]
            X = jnp.array(X)
            protein_exp = jnp.array(protein_exp)
            sample_batch = jnp.array(sample_batch)

            d_rng, _ = jax.random.split(d_rng)
            loss, state, batch_stats = train_step(
                state, batch_stats, [X, sample_batch, protein_exp], d_rng
            )

        val_loss = 0
        n_batches = 0
        for batch in val_dl:
            X = batch[REGISTRY_KEYS.X_KEY]
            protein_exp = batch["response"]
            sample_batch = batch[REGISTRY_KEYS.BATCH_KEY]

            X = jnp.array(X)
            protein_exp = jnp.array(protein_exp)
            sample_batch = jnp.array(sample_batch)

            _val_loss = evaluate(state, batch_stats, [X, sample_batch, protein_exp])
            n_batches += 1
            val_loss += _val_loss

        losses.append(
            dict(
                metric="stat_train_loss",
                value=loss.item(),
                iterate=epoch,
            )
        )
        losses.append(
            dict(
                metric="stat_val_loss",
                value=val_loss.item() / n_batches,
                iterate=epoch,
            )
        )

        if val_loss < best_val_loss:
            best_val_loss = val_loss
            hasnt_improved = 0
            best_params = copy(state.params)
            best_batch_stats = copy(batch_stats)
        else:
            hasnt_improved += 1
        if hasnt_improved > patience:
            break
    losses = pd.DataFrame(losses)
    return losses, best_params, best_batch_stats


def evaluate_statistic(
    statistic_model,
    params,
    batch_stats,
    val_adata,
    batch_key,
    protein_key,
    batch_size,
    normalize_x=True,
    include_batch_in_input=False,
):
    """Computes the validation loss of the feature importance model."""

    n_samples = val_adata.obs[batch_key].nunique()

    @jax.jit
    def evaluate(inputs):
        x, y = inputs
        x_ = (
            jnp.log1p(1e6 * x / jnp.sum(x, axis=-1, keepdims=True))
            if normalize_x
            else x
        )
        x_ = (
            jnp.concatenate(
                [x_, jax.nn.one_hot(batch_sample.squeeze(-1), n_samples)], axis=-1
            )
            if include_batch_in_input
            else x_
        )
        loss = statistic_model.apply(
            {"params": params, "batch_stats": batch_stats},
            x_,
            y,
            training=False,
        )["loss"]
        return loss

    val_dl = construct_dataloader(
        val_adata,
        batch_size=batch_size,
        shuffle=False,
        batch_key=batch_key,
        protein_key=protein_key,
    )

    val_loss = 0
    n_batches = 0
    for batch in val_dl:
        X = batch[REGISTRY_KEYS.X_KEY]
        protein_exp = batch["response"]
        batch_sample = batch[REGISTRY_KEYS.BATCH_KEY]

        X = jnp.array(X)
        protein_exp = jnp.array(protein_exp)
        batch_sample = jnp.array(batch_sample)

        _val_loss = evaluate([X, batch_sample, protein_exp])
        n_batches += 1
        val_loss += _val_loss
    return val_loss.item() / n_batches
